@@ -7,9 +7,11 @@ module Main where
 
 import           Conduit
 import           Control.Exception
-import           Control.Monad         (void, when, zipWithM_)
+import           Control.Monad         (forM_, void, when, zipWithM_)
 import           Data.ByteString       as S
 import           Data.ByteString.Char8 as S8
+import qualified Data.ByteString.Lazy.Char8 as LS8
+import           Data.ByteString.Short (fromShort)
 import           Data.Conduit.List
 import           Data.Conduit.Tar
 import           Data.Int
@@ -38,10 +40,21 @@ main = do
     let baseTmp = "tar-conduit-tests"
     isStack <- doesDirectoryExist ".stack-work"
     let testPaths =
-            ["src", "./tests", "README.md", "ChangeLog.md", "LICENSE"] ++
-            if isStack
-                then [".stack-work"]
-                else []
+               ["src", "README.md", "ChangeLog.md", "LICENSE"]
+#ifndef WINDOWS
+            <> ["./tests"]
+               -- On Windows, the 'stack test' command results in error message:
+               --
+               --   uncaught exception: IOException of type PermissionDenied
+               --   System.Win32File.read: permission denied (Permission denied)
+               --
+               -- if '.stack-work' is included in the test paths.
+            <> [".stack-work" | isStack]
+#else
+               -- The package does not support symlinks on Windows. See
+               -- Data.Conduit.Tar.Windows.getFileInfo.
+            <> ["./tests/files"]
+#endif
     hspec $ do
         describe "tar/untar" $ do
             let tarUntarContent dir =
@@ -74,6 +87,7 @@ main = do
                     collectContent (outDir Posix.</> "dir/subdir/") `shouldReturn` "Hello World\n"
         describe "ustar" ustarSpec
         describe "GNUtar" gnutarSpec
+        describe "pax" paxSpec
         describe "unsupported headers" $ do
             it "associated payload is discarded" $ do
               contents <- readGzipTarball "./tests/files/libpq-0.3.tar.gz"
@@ -278,3 +292,145 @@ collectContent :: FilePath -> IO (ByteString)
 collectContent dir =
     runConduitRes $
     sourceDirectoryDeep False dir .| mapMC (\fp -> runConduit (sourceFileBS fp .| foldC)) .| foldC
+
+-- | This test uses untar to process a simple example in the pax interchange
+-- format.
+paxSpec :: Spec
+paxSpec = assert payloadSizeCheck $ do
+    it "untarChunksRaw, pax interchange format" $ do
+        res <- runConduitRes $
+               paxExample
+            .| chopEvery bigChopSize
+            .| untarChunksRaw
+            .| processTarChunks
+            .| sinkList
+        pure res `shouldReturn`
+            [ "/pax-global-header"
+            , "payload: 19 comment=Example\n"
+            , "/pax-extended-header"
+            , "payload: " <> bigPayload1
+            , "payload: " <> bigPayload2
+            , "original-dir/original-filepath"
+            , "payload: " <> smallPayload
+            ]
+    it "untar, pax interchange format" $ do
+        res <- runConduitRes $
+               paxExample
+            .| chopEvery smallChopSize
+            .| untar process
+            .| sinkList
+        pure res `shouldReturn`
+            [ (veryLongFilepath, smallPayload1)
+            , (veryLongFilepath, smallPayload2)
+            ]
+    it "untarRaw, pax interchange format" $ do
+        res <- runConduitRes $
+               paxExample
+            .| chopEvery smallChopSize
+            .| untarRaw process
+            .| sinkList
+        pure res `shouldReturn`
+            [ ("original-dir/original-filepath", smallPayload1)
+            , ("original-dir/original-filepath", smallPayload2)
+            ]
+  where
+    process fi = awaitForever $ \bs -> yield (filePath fi, bs)
+    processTarChunks = awaitForever $ \tc -> yield $ case tc of
+        ChunkHeader h -> fromShort $
+            headerFileNamePrefix h <> "/" <> headerFileNameSuffix h
+        ChunkPayload _ bs -> "payload: " <> bs
+        ChunkException e -> "exception: " <> S8.pack (show e)
+    chopEvery :: (MonadIO m) => Int -> ConduitT ByteString ByteString m ()
+    chopEvery n = chop
+      where
+        chop = await >>= \case
+            Nothing -> pure ()
+            Just val -> do
+                forM_ (split (S.unpack val)) $ \chunk -> yield (S.pack chunk)
+                chop
+        split = P.takeWhile (not . P.null)
+              . P.map (P.take n)
+              . P.iterate (P.drop n)
+    bigChopSize = 512
+    (bigPayload1, bigPayload2) = S.splitAt bigChopSize veryLongFilepathRecord
+    smallChopSize = 4
+    (smallPayload1, smallPayload2) = S.splitAt smallChopSize smallPayload
+    moreThanHalf s l = l > s && l <= 2 * s
+    payloadSizeCheck =
+           moreThanHalf bigChopSize (S.length veryLongFilepathRecord)
+        && moreThanHalf smallChopSize (S.length smallPayload)
+
+
+-- | Produces a simple example in the pax interchange format. It has a pax
+-- \'global\' header block providing a comment, a pax \'next\' header block
+-- providing a very long path (@\"very\very\...\very\long\filepath\"@), and a
+-- normal file with filepath @\"original-filepath\"@ and payload @\"payload\".
+paxExample :: MonadThrow m => ConduitM a ByteString m ()
+paxExample = void $
+       yieldMany
+           [ Left globalHeader
+           , Right globalPayload
+           , Left extendedHeader
+           , Right extendedPayload
+           , Left ustarHeader
+           , Right ustarPayload
+           ]
+    .| void tarEntries
+  where
+    defaultHeader :: FileOffset -> Header
+    defaultHeader offset = Header
+        { headerOffset = offset
+        , headerPayloadOffset = offset + 512
+        , headerFileNameSuffix = mempty
+        , headerFileMode = 0o666
+        , headerOwnerId = 0x0
+        , headerGroupId = 0x0
+        , headerPayloadSize = 0x0
+        , headerTime = 0x0
+        , headerLinkIndicator = 0x0
+        , headerLinkName = mempty
+        , headerMagicVersion = "ustar"
+        , headerOwnerName = "root"
+        , headerGroupName = "root"
+        , headerDeviceMajor = 0x0
+        , headerDeviceMinor = 0x0
+        , headerFileNamePrefix = mempty
+        }
+    nextOffset :: Header -> FileOffset
+    nextOffset h =
+        let payloadRecordCount = (headerPayloadSize h + 511) `div` 512
+        in  headerPayloadOffset h + 512 + payloadRecordCount * 512
+    globalHeader = (defaultHeader 0x0)
+        { headerFileNameSuffix = "pax-global-header"
+        , headerPayloadSize = fromIntegral $ S.length globalPayload
+        , headerLinkIndicator = 0x67 -- UTF-8 'g'
+        }
+    globalPayload = "19 comment=Example\n"
+    extendedHeader = (defaultHeader $ nextOffset globalHeader)
+        { headerFileNameSuffix = "pax-extended-header"
+        , headerPayloadSize = fromIntegral $ S.length extendedPayload
+        , headerLinkIndicator = 0x78 -- UTF-8 'x'
+        }
+    -- The path in the pax extended header should override the filepath
+    -- specified in the ustar header.
+    extendedPayload = veryLongFilepathRecord
+    ustarHeader = (defaultHeader $ nextOffset extendedHeader)
+        { headerFileNameSuffix = "original-filepath"
+        , headerPayloadSize = fromIntegral $ S.length ustarPayload
+        , headerLinkIndicator = 0x30 -- UTF-8 '0'
+        , headerFileNamePrefix = "original-dir"
+        }
+    ustarPayload = smallPayload
+
+-- | A very/very/.../very/long/filepath with 653 bytes.
+veryLongFilepath :: ByteString
+veryLongFilepath =
+    S8.toStrict (LS8.take 640 $ LS8.cycle "very/") <> "long/filepath"
+
+-- | A very, very, ..., very, long filepath record with 663 bytes.
+veryLongFilepathRecord :: ByteString
+veryLongFilepathRecord = "663 path=" <> veryLongFilepath <> "\n"
+
+-- | A small payload.
+smallPayload :: ByteString
+smallPayload = "payload"
